/*
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <error.h>
#include <ios1.h>
#include <proto/dos.h>

#define TRUE 1
#define FALSE 0

#define WARN 10
#define ERROR 20

#define MAXLEN 255 /* Max line length */
#define MAXDELIM 20
#define MAXLIST 260 /* > MAXLEN */

#define isoctal(c) ((c) >= '0' && (c) <= '7')
#define hexdigit(c) ((c) <= '9' ? (c) - '0' : tolower(c) - 'a')

extern char *optarg;
extern int opterr, optind;
extern int _bufsiz;
int err;

char delims[MAXDELIM] = "\t ";
int suppress, column, field, delim, unique;
char use[MAXLIST];

int nextfield(char *field, char *line, char **pos)
{
    static int done;
    char *end;

    if (*pos == NULL)
    {
        if (unique)
            *pos = line;
        else
            *pos = line + strspn(line, delims);
        done = FALSE;
    }
    if (done) return EOF;

    end = strpbrk(*pos, delims);

    if (!end)
    {
        strcpy(field, *pos);
        done = TRUE;

        return '\0';
    }
    else
    {
        field[end - *pos] = '\0';
        strncpy(field, *pos, end - *pos);
        if (unique)
            *pos = end + 1;
        else
            *pos = end + strspn(end, delims);

        return *end;
    }
}

char *extract_list(char *to, char *from)
{
    char *ret = to;
    int pos = 0;

    while (pos < MAXLIST - 1 && from[pos])
    {
        if (use[pos + 1])
           *to++ = from[pos];
        pos++;
    }
    *to++ = '\n'; /* Shouldn't be here, but ... */
    *to = '\0';
    return ret;
}

char *expand_tabs(char *dest, char *src, int maxlen)
{
    int spos = 0, dpos = 0;

    while (src[spos] && dpos < maxlen - 1)
    {
        if (src[spos] != '\t') dest[dpos++] = src[spos++];
        else
        {
            int pad = 8 - spos % 8;

            if (dpos + pad <= maxlen - 1)
            {
                memset(&dest[dpos], ' ', 8 - spos % 8);
                dpos += pad;
                spos++;
            }
            else break;
        }
    }
    dest[dpos] = '\0';
    return dest;
}

void cut(FILE *file)
{
    char line[MAXLEN + 1], detabbed[MAXLEN + 1], out[MAXLEN + 2];

    line[MAXLEN] = '\0';

    while (errno == 0 && fgets(line, MAXLEN, file) && errno == 0)
    {
        int l = strlen(line);

        if (l > 0 && line[l - 1] == '\n') line[l - 1] = '\0';

        if (column)
        {
            expand_tabs(detabbed, line, MAXLEN + 1);
            extract_list(out, detabbed);
            fputs(out, stdout);
        }
        else
            if (strpbrk(line, delims) != NULL)
            {
                int pos = 1, del, lastdel = 0;
                char *next = NULL;

                while (errno == 0 && pos < MAXLIST && (del = nextfield(out, lin
e, &next)) != EOF)
                {
                    if (use[pos])
                    {
                        if (lastdel) fputc(lastdel, stdout);
                        lastdel = del;
                        if (errno == 0) fputs(out, stdout);
                    }
                    pos++;
                }
                if (errno == 0) fputc('\n', stdout);
            }
            else if (!suppress)
            {
                fputs(line, stdout);
                fputc('\n', stdout);
            }
    }
    if (errno != 0)
    {
        perror("cut failed");
        err = ERROR;
    }
}

char *c_str(char *to, char *from, int cnt)
{
    char *ret = to;

    while (from[0] && --cnt)
    {
        if (from[0] == '\\' && from[1])
        {
            switch (*++from)
            {
                case 'n':
                    *to++ = '\n';
                    break;
                case 'r':
                    *to++ = '\r';
                    break;
                case 't':
                    *to++ = '\t';
                    break;
                case 'b':
                    *to++ = '\b';
                    break;
                case 'f':
                    *to++ = '\f';
                    break;
                case 'x':
                    if (isxdigit(from[1]) && isxdigit(from[2]))
                    {
                        *to++ = hexdigit(from[1]) << 4 | hexdigit(from[2]);
                        from += 2;
                    }
                    else
                        fputs("Bad hex char\n", stderr);
                    break;
                case '0': case '1': case '2': case '3':
                case '4': case '5': case '6': case '7':
                    if (isoctal(from[1]) && isoctal(from[2]))
                    {
                        *to++ = (from[0] - '0') << 6 | (from[1] - '0') << 3 |
                                (from[2] - '0');
                        from += 2;
                    }
                    else
                        fputs("Bad octal char\n", stderr);
                    break;
                default:
                    *to++ = *from;
                    break;
            }
            from++;
        }
        else
            *to++ = *from++;
    }
    *to = '\0';
    return ret;
}

int analyse_range(char *range)
{
    int last, first;

    range = stpblk(range);

    if (*range == '-') first = 1;
    else
    {
        if (!isdigit(*range)) return FALSE;
        range += stcd_i(range, &first);
        range = stpblk(range);
    }
    if (*range == '-')
    {
        range++;
        if (isdigit(*range))
            range += stcd_i(range, &last);

        else
            last = MAXLIST - 1;
    }
    else last = first;

    if (*stpblk(range) == '\0' && first > 0 && first <= last && last < MAXLIST)
     
    {
        int i;

        for (i = first; i <= last; i++) use[i] = TRUE;

        return TRUE;
    }
    else
        return FALSE;
}

int analyse_list(char *list)
{
    int i;
    char *end;

    for (i = 1; i < MAXLIST; i++) use[i] = FALSE;

    while (end = strchr(list, ','))
    {
        *end = '\0';
        if (!analyse_range(list)) return FALSE;
        list = end + 1;
    }
    return analyse_range(list);
}

void main(int argc, char **argv)
{
    int usage = TRUE, ch;

    /* Setup stdout */
    if (IsInteractive(chkufb(fileno(stdout))->ufbfh))
        setvbuf(stdout, NULL, _IOLBF, _bufsiz);

    /* Analyse command line */
    while (usage && (ch = getoptx(argc, argv, "c:d:f:su")) != EOF)
    {
        switch (ch)
        {
            case 'c':
                column = TRUE;
                usage = analyse_list(optarg);
                break;
            case 'd':
                delim = TRUE;
                c_str(delims, optarg, MAXDELIM);
                break;
            case 'f':
                field = TRUE;
                usage = analyse_list(optarg);
                break;
            case 's':
                suppress = TRUE;
                break;
            case 'u':
                unique = TRUE;
                break;
            default:
                usage = FALSE;
                break;
        }
    }
    if (!(column && !(delim || field || suppress) || field && !column))
        usage = FALSE;

    if (!usage)
    {
        fprintf(stderr, "Usage:\t%s -clist [files]\n\t%s -flist [-dchar] [-su]
[files]\n", argv[0], argv[0]);
        err = WARN;
    }
    else
        /* Do cut */
        if (argc == optind)
            cut(stdin);
        else
        {
            int i;

            for (i = optind; err < ERROR && i < argc; i++)
            {
                FILE *f = fopen(argv[i], "r");

                if (!f)
                {
                    fprintf(stderr, "Error opening %s", argv[i]);
                    perror("");
                    err = WARN;
                }
                else
                {
                    cut(f);
                    fclose(f);
                }
            }
        }
    exit(err);
}
